#ifndef NVLISTS_H
#define NVLISTS_H

#include "defs.h"

extern const struct name_value service_types[];
extern const struct name_value service_flags[];
extern const struct name_value socket_types[];
extern const struct name_value success_log_options[];
extern const struct name_value failure_log_options[];
extern const struct name_value syslog_facilities[];
extern const struct name_value syslog_levels[];

#endif
